libfte-experimental
===================

**This is experimental software and is not suitable for production environments.**

This is an experimental branch of [libfte](https://github.com/kpdyer/libfte), implementing the concepts from:

LibFTE: A Toolkit for Constructing Practical, Format-Abiding Encryption Schemes   
Daniel Luchaup, Kevin P. Dyer, Somesh Jha, Thomas Ristenpart and Thomas Shrimpton    
To appear in the proceedings of USENIX Security 2014

Building
--------

**We currently support Ubuntu 14.04**

```console
$ sudo apt-get install git build-essential bison flex libboost-python-dev libgmp-dev python-crypto python-gmpy
$ git clone https://github.com/kpdyer/libfte-experimental.git
$ cd libfte-experimental
$ make
```

Testing
-------

```console
$ make test
['strategy=0', "input=('(a|b)+b(a|b)+', 16)", "ouput=('(a|b)+b(a|b)+', 16)"]
['strategy=0', "input=('(a|b)+b(a|b)+', 16)", "ouput=('(a|b)+b(a|b)+', 32)"]
...
```

Dependencies
------------

* build tools: make/gcc/g++/etc.
* Python 2.7
* Boost.python
* Bison
* Flex
* GMP
* gmpy
* PyCrypto

Examples
--------
We have two scripts in the ```examples/``` directory that print a sample of what our plaintext/ciphertext look like for some input/ouput format pairs.

First we have an FPE scheme with input/output as a 30-bit string.

```console
$ python bin2bin.py 
input_format=('(0|1)+', 30)
output_format=('(0|1)+', 30)
plaintext=010101000101010010111001000010
ciphertext=000001111010111111101010110100
```

```console
$ python bin2bin.py 
input_format=('(0|1)+', 30)
output_format=('(0|1)+', 30)
plaintext=101001000101011000101011010110
ciphertext=111011000000110101101110111111
```

Next we have input of 13-bit strings and 5-character hex strings.

```console
$ python bin2hex.py 
input_format=('(0|1)+', 13)
output_format=('[0-9A-F]+', 5)
plaintext=0000011010100
ciphertext=8D07A
```

```console
$ python bin2hex.py 
input_format=('(0|1)+', 13)
output_format=('[0-9A-F]+', 5)
plaintext=0011111111100
ciphertext=55F57
```

As another example I have as input a regular expression representing something like an email address, as output a regular expression representing something ilke a URL.

```console
$ python email2url.py 
input_format=('^([a-z0-9_\\.]+)@([\\da-z\\.-]+)\\.([a-z\\.]{2,6})$', 32)
output_format=('^(https?:\\/\\/)([\\da-z\\.-]+)\\.([a-z\\.]{2,6})([\\/\\w \\.-]*)*\\/?$', 64)
plaintext=tfrzt89soztefz92mo0jikli_2@7t.un
ciphertext=http://li.crFW.2Hy0gEQ6ttELwaZEf.go4OcjSGIqQFaOGVEoyjzZPjGy-HZ0H
```

```console
$ python email2url.py 
input_format=('^([a-z0-9_\\.]+)@([\\da-z\\.-]+)\\.([a-z\\.]{2,6})$', 32)
output_format=('^(https?:\\/\\/)([\\da-z\\.-]+)\\.([a-z\\.]{2,6})([\\/\\w \\.-]*)*\\/?$', 64)
plaintext=775d7e_9wtc1p4amh3we3@9rgjsv..ez
ciphertext=http://y6un..sm0yFDic2lRn0Q6HPR.nXXvUVoLZVUWjI_ZRp_GQ_zlR2OE34GE
```

Authors
=======

* Kevin P. Dyer (FTE and thirdparty/ffx)
* Daniel Luchaup (thirdparty/ranking-test-framework)
