ranking
=======

Tested on Ubuntu 13.10.

Overview
--------

Ranking strategies

* ```gs_luchaup``` - traditional GS implemented by Daniel
* ```gs_plus_luchaup``` - an algorithmicly-improved version of GS, implemented by Daniel
* ```nfa_luchaup``` - an NFA-based ranking algorithm, implemented by Daniel
* ```gs_dyer``` - traditional GS, implemented by Kevin
* ```gs_plus_dyer``` - an algorithmicly-improved version of GS, implemented by Kevin

The command ```make all``` produces two static libraries: ```src/fteproxy/rank_unrank.a``` and ```src/luchaup/rank.a```. These libraries are used by ```src/performance_benchmark.cc``` to produce ```bin/performance_benchmark```, which is a program that output CSV-formatted performance metrics about the various (un)ranking strategies.

Dependencies
------------

* standard build tools: gcc/g++/make/etc.
* openfst: http://www.openfst.org/
* bison: http://www.gnu.org/software/bison/
* flex: http://flex.sourceforge.net/
* GMP: https://gmplib.org/

Install OpenFST from source.
Install GMP/bison/flex using something like: ```sudo apt-get install libgmp-dev bison flex```


Building
--------
Install depdencies, then:
```
$ make all
[snip]
$ ./bin/performance_benchmkark --help
usage: performance_benchmark [strategy] [algorithm] [regex] [max_len]
  strategy = gs_luchaup | gs_plus_luchaup | nfa_luchaup | gs_dyer | gs_plus_dyer
  algorithm = memoization | rank | unrank
  regex = [user defined regular expression R]
  max_len = maximum string length in L(R) to compute for memoization
```

Structure
---------

* ```bin/performance_benchmark``` - a program that tests the various (un)ranking strategy and produces a CSV-formatted report
* ```bin/make_graphs``` - given CSV output from bin/performance_benchmark, produces PDF graphs
* ```issues/``` - contains code snippets for bugs identified in src
* ```src/``` - contains all source code for (un)ranking from fteproxy and Daniel Luchaup's implementation

Examples
--------

We have the following enum which describes the different ranking strategies, as described above in overview.
```
regex_ranker_strategy::DFA_RANKING1
regex_ranker_strategy::DFA_RANKING2
regex_ranker_strategy::NFA_RANKING
```

We can then initialize a regex_ranker object and use it for (un)ranking.
```
#include "regex_ranker.h"
...

// initialize the object
size_t strategy = regex_ranker_strategy::DFA_RANKING1;
std::string regex = "(0|1)+";
size_t max_len = 8;
regex_ranker rankerObj(strategy, regex, max_len);

// use the generated object to (un)rank
std::string = rankerObj.unrank(0);
mpz_class C = rankerObj.rank("00000000");
```
