/*-----------------------------------------------------------------------------
 * File:    ex2G_defs.cc
 * Author:  Daniel Luchaup
 * Date:    December 2013
 * Copyright 2013 Daniel Luchaup luchaup@cs.wisc.edu
 *
 * This file contains unpublished confidential proprietary work of
 * Daniel Luchaup, Department of Computer Sciences, University of
 * Wisconsin--Madison.  No use of any sort, including execution, modification,
 * copying, storage, distribution, or reverse engineering is permitted without
 * the express written consent (for each kind of use) of Daniel Luchaup.
 *
 *---------------------------------------------------------------------------*/
#include <string.h>
#include "ex2G_defs.h"
#include "ffa.h"

/* This is grammar G:
 * *** grammar rules ***
 * r0: S  -> ( Exp )
 * r1: S  -> ( Exp + S)
 * r2: Exp -> IDENTIFIER
 * r3: Exp -> Num
 * r4: Num -> INT
 * r5: Num -> FLOAT
 * *** lexical rules ***
 * r6: IDENTIFIER -> {L}({L}|{D})*
 * r7: INT -> 0[xX]{H}+
 * r8: INT -> {D}+
 * r9: FLOAT -> {D}*"."{D}+
 * r10:FLOAT -> {D}+"."{D}*
 */
Grammar G;

/************* Terminals ***************/
Terminal *lp;
Terminal *rp;
Terminal *plus;

/********  normal Nonterminals *********/
Nonterminal *L;
Nonterminal *S;
Nonterminal *Exp;
/******** lexical Nonterminals *********/
Nonterminal *Num;
Nonterminal *IDENTIFIER;
Nonterminal *INT;
Nonterminal *FLOAT;

/********  normal Rules *********/
GRule *r00;//r00: L  -> S
GRule *r01;//r01: L  -> S L
GRule *r0; //r0: S  -> ( Exp )
GRule *r1; //r1: S  -> ( Exp + S)
GRule *r2; //r2: Exp -> IDENTIFIER
GRule *r3; //r3: Exp -> Num
GRule *r4; //r4: Num -> INT
GRule *r5; //r5: Num -> FLOAT
/******** lexical Rules *********/
LRule *r6; //r6: IDENTIFIER -> {L}({L}|{D})*
LRule *r7; //r7: INT -> 0[xX]{H}+
LRule *r8; //r8: INT -> {D}+
LRule *r9; //r9: FLOAT -> {D}*"."{D}+
LRule *r10;//r10: FLOAT -> {D}+"."{D}*

void init_G() {
  /*** Here is how to build a representation, G, for the above grammar ***/  
  lp = new Terminal("(");
  rp = new Terminal(")");
  plus = new Terminal("+");
  
  L = new Nonterminal("L");
  S = new Nonterminal("S");
  Exp = new Nonterminal("Exp");
  Num = new Nonterminal("Num");
  IDENTIFIER = new Nonterminal("IDENTIFIER");
  INT = new Nonterminal("INT");
  FLOAT = new Nonterminal("FLOAT");

  G.terminals.push_back(lp);
  G.terminals.push_back(rp);
  G.terminals.push_back(plus);
    
  G.nonterminals.push_back(L); G.index_start = 0;
  G.nonterminals.push_back(S); //G.index_start = 0;
  G.nonterminals.push_back(Exp);
  G.nonterminals.push_back(Num);
  G.nonterminals.push_back(IDENTIFIER);
  G.nonterminals.push_back(INT);
  G.nonterminals.push_back(FLOAT);
  
  r00 = new GRule(L, "r00");
  r00->append_symbol(S);
  
  r01 = new GRule(L, "r01");
  r01->append_symbol(S); r01->append_symbol(L);
  
  r0 = new GRule(S, "r0");
  r0->append_symbol(lp);r0->append_symbol(Exp);r0->append_symbol(rp);
  
  r1 = new GRule(S, "r1");
  r1->append_symbol(lp);r1->append_symbol(Exp);r1->append_symbol(plus);r1->append_symbol(S);r1->append_symbol(rp);
  
  r2 = new GRule(Exp, "r2");
  r2->append_symbol(IDENTIFIER);
  
  r3 = new GRule(Exp, "r3");
  r3->append_symbol(Num);
  
  r4 = new GRule(Num, "r4");
  r4->append_symbol(INT);

  r5 = new GRule(Num, "r5");
  r5->append_symbol(FLOAT);

  // NOTE: we replaced the lex macros {D},{L},{H}. Ex: replaced {D} with [0-9]
  unsigned int lim = 3;
  r6 = new LRule(IDENTIFIER, new rank_ffa_tab_t("/[a-zA-Z_]([a-zA-Z_]|[0-9])*/"), "r6", 2);
  r7 = new LRule(INT, new rank_ffa_tab_t("/0[xX][a-fA-F0-9]+/"), "r7", lim);
  r8 = new LRule(INT, new rank_ffa_tab_t("/[0-9]+/"), "r8", lim);
#if 1 //ambiguous
  r9 = new LRule(FLOAT, new rank_ffa_tab_t("/[0-9]*\\.[0-9]+/"), "r9", lim);
  r10 = new LRule(FLOAT, new rank_ffa_tab_t("/[0-9]+\\.[0-9]*/"), "r10", lim);
#else //nonambiguous
  r9 = new LRule(FLOAT, new rank_ffa_tab_t("/([0-9]*\\.[0-9]+)|([0-9]+\\.[0-9]*)/"), "r9", lim);
#endif
}

void init_G(unsigned int slice_size) {
  init_G();
  (const_cast<Nonterminal*>(G.S()))->memoize(slice_size, slice_size);
}

///////////// grammar parsing stuff ///////////////////
// copied from Randy's parse_utils. TBD: refactor this.
typedef enum {
	EVAL_SRC_FILE,
	EVAL_SRC_MEM
} eval_source;


/* Ugly, but it gets the job done... */
static struct {
	eval_source type;
	union {
		FILE *file;
		struct {
			const char *data;
			int len;
		} mem;
	} src;
} eval_data;// TBD use one of this per yacc file, to allow multilpe parsers

extern ParseTree *g_ex2G_ParseTree;

ParseTree *pt_from_file(FILE *fp) //TBD: move in shared file
{
	eval_data.type = EVAL_SRC_FILE;
	eval_data.src.file = fp;
	//lineno=1;
	yyparse();
	return g_ex2G_ParseTree;
}

ParseTree *pt_from_mem(const char *data, int len) //TBD: move in shared file
{
	eval_data.type = EVAL_SRC_MEM;
	eval_data.src.mem.data = data;
	eval_data.src.mem.len = len;
	//lineno = line;
	yyparse();
	return g_ex2G_ParseTree;
}

ParseTree *pt_from_mem(const std::string &data) //TBD: move in shared file
{
  return pt_from_mem(data.c_str(), data.length());
}

int switch_getinput(char *buf, int maxlen)
{
   int retval = 0;
   
   switch (eval_data.type) {
      case EVAL_SRC_FILE:
	 retval = fread(buf, 1, maxlen, eval_data.src.file);
	 break;
      case EVAL_SRC_MEM:
	 if ( maxlen > eval_data.src.mem.len ) {
	    maxlen = eval_data.src.mem.len;
	 }
	 memcpy(buf, eval_data.src.mem.data, maxlen);
	 eval_data.src.mem.data += maxlen;
	 eval_data.src.mem.len -= maxlen;
	 retval = maxlen;
	 break;
   }
   return retval;
}
