/*-----------------------------------------------------------------------------
 * File:    ex2G_defs.h
 * Author:  Daniel Luchaup
 * Date:    December 2013
 * Copyright 2013 Daniel Luchaup luchaup@cs.wisc.edu
 *
 * This file contains unpublished confidential proprietary work of
 * Daniel Luchaup, Department of Computer Sciences, University of
 * Wisconsin--Madison.  No use of any sort, including execution, modification,
 * copying, storage, distribution, or reverse engineering is permitted without
 * the express written consent (for each kind of use) of Daniel Luchaup.
 *
 *---------------------------------------------------------------------------*/
#ifndef _EX2G_DEFS_
#define _EX2G_DEFS_
#include "gram.h"
#include <string>

/* This is grammar G:
 * *** grammar rules ***
 * r0: S  -> ( Exp )
 * r1: S  -> ( Exp + S)
 * r2: Exp -> IDENTIFIER
 * r3: Exp -> Num
 * r4: Num -> INT
 * r5: Num -> FLOAT
 * *** lexical rules ***
 * r6: IDENTIFIER -> {L}({L}|{D})*
 * r7: INT -> 0[xX]{H}+
 * r8: INT -> {D}+
 * r9: FLOAT -> {D}*"."{D}+
 * r10:FLOAT -> {D}+"."{D}*
 */
extern Grammar G;

/************* Terminals ***************/
extern Terminal *lp;
extern Terminal *rp;
extern Terminal *plus;

/********  normal Nonterminals *********/
extern Nonterminal *L;
extern Nonterminal *S;
extern Nonterminal *Exp;
/******** lexical Nonterminals *********/
extern Nonterminal *Num;
extern Nonterminal *IDENTIFIER;
extern Nonterminal *INT;
extern Nonterminal *FLOAT;

/********  normal Rules *********/
extern GRule *r00;//r00: L  -> S
extern GRule *r01;//r01: L  -> S L
extern GRule *r0; //r0: S  -> ( Exp )
extern GRule *r1; //r1: S  -> ( Exp + S)
extern GRule *r2; //r2: Exp -> IDENTIFIER
extern GRule *r3; //r3: Exp -> Num
extern GRule *r4; //r4: Num -> INT
extern GRule *r5; //r5: Num -> FLOAT
/******** lexical Rules *********/
extern LRule *r6; //r6: IDENTIFIER -> {L}({L}|{D})*
extern LRule *r7; //r7: INT -> 0[xX]{H}+
extern LRule *r8; //r8: INT -> {D}+
extern LRule *r9; //r9: FLOAT -> {D}*"."{D}+
extern LRule *r10;//r10: FLOAT -> {D}+"."{D}*

extern void init_G();
extern void init_G(unsigned int slice_size);
extern void deallocate_G();

#define yyparse ex2G_parse
#define yylval ex2G_lval
extern int yyparse();
extern ParseTree *g_ex2G_ParseTree;

///////////// grammar parsing stuff ///////////////////
ParseTree *pt_from_file(FILE *fp);
ParseTree *pt_from_mem(const char *data, int len);
ParseTree *pt_from_mem(const std::string &data);

#endif
