/*-----------------------------------------------------------------------------
 * File:    rank.cc
 * Author:  Daniel Luchaup
 * Date:    26 November 2013
 * Copyright 2010-2013 Daniel Luchaup luchaup@cs.wisc.edu
 *
 * This file contains unpublished confidential proprietary work of
 * Daniel Luchaup, Department of Computer Sciences, University of
 * Wisconsin--Madison.  No use of any sort, including execution, modification,
 * copying, storage, distribution, or reverse engineering is permitted without
 * the express written consent (for each kind of use) of Daniel Luchaup.
 *
 *---------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <vector>
#include <iostream>
#include <sstream>
#include <algorithm>
#include <assert.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <errno.h>
#include <assert.h>
#include <pthread.h>

#include "zoptions.h"
#include "dbg.h"
#include "MyTime.h"
#include "dfa_tab_t.h" //for minimMDFA : TBD: may want to get rid of it...

static std::string action;

/*-----------------------------------------------------------------------------
 *get_options
 *---------------------------------------------------------------------------*/
zOptions zop(//The following is the usage message
"(NOTE!: this is mostly obsolete... TBD: update it!)\n\
  -a <action> : action to perform\n\
\t Must specify exactly one action from\n\
  -z : pass extra options\n\
  Ex.:\n\
    -z sigFile:FILE read the signatures mdfa from FILE \n\
    -z MDFASTATES:MAX Use at most MAX number of states in the mdfa.\n\
");

void get_options(int argc, char *argv[])
{
  int c;
  extern char *optarg;
  
  while ((c = getopt(argc, argv, "a:z:T:")) != -1) {
    switch (c)
      {
      case 'a':
	if (!action.empty())
	  zop.usage_and_die("dupplicate 'action' argument");
	action = optarg;
	break;
      case 'z':
	zop.dbgOptionSet.insert(std::string(optarg));
	break;
      default:
	zop.usage_and_die("unknown option");
	break;
      }
  }
  
  return;
}
///////////////////////////////////////////////////////////////////////////////
void init(int argc, char *argv[]) {
  // Debug: dump the arguments
  std::cout << "Command line: ";
  for (int i = 0; i < argc; ++i) {
    std::cout << argv[i] << " ";
  }
  std::cout << std::endl;
#ifndef SVN_VERSION
#define SVN_VERSION "UNKNOWN"
#endif

#ifndef SVN_CHANGED
#define SVN_CHANGED "UNCHANGED"
#endif

  DMPNL(SVN_VERSION);
  DMPNL(SVN_CHANGED);
  
  minimMDFA = zop.dbgNumOption("minimMDFA",0) == 1;
  std::cout << "action = " << action << std::endl;
  zop.dumpOptions();
  DMP(minimMDFA) << std::endl;
  DMP(zop.dbgDoubleOption("falsePositives", -1)) << std::endl;
  if (action.empty())
    zop.usage_and_die("missing 'action' argument");
  if (zop.dbgDefinedOption("cutLim"))
    zop.usage_and_die("Obsolete option 'cutLim'. Use 'falsePositives' instead.");
}

void main_action() {
  if (action == "trex_tests") {
    // rank -atrex_tests -zrexExprFile:<REX-in-file> -zrexInputFile:<string-input-file>
    // Use TREX_EXPAND_DOTS to expand '.' in regular expressions to OR of all characters
    extern void trex_tests();
    trex_tests();
  }
  else if (action == "rank_tests") {
    // Use TREX_EXPAND_DOTS to expand '.' in regular expressions to OR of all characters
    // rank -arank_tests -zrexExprFile:TEST_REXs -zallow_ambiguous -zforced_len:20
    // rank -arank_tests -zrexExprFile:TEST_REXs [-zallow_ambiguous] [-zforced_len:20] -z(do_dfa|do_nfa|do_trx|do_ffx) [-zrepeat:100]
    extern void rank_tests();
    rank_tests();
  }
  else if (action == "unit_test") {
    // rank -aunit_test -zrexExprFile:TEST_REXs [-zrexInputFile:<string-input-file>] [-zallow_ambiguous] [-zforced_len:20] -z(do_dfa|do_nfa|do_trx|do_ffx) [-zrepeat:100]
    extern void unit_test();
    unit_test();
  }
  else if (action == "rex_classify") {
    // rank -arex_classify -zrexExprFile:TEST_REXs -zcheck_ambiguity -zforced_len:20
    extern void rex_classify();
    rex_classify();
  }
  else if (action == "rex_nfa_classify") {
    // rank -arex_nfa_classify -zrexExprFile:TEST_REXs -zforced_len:20
    extern void rex_nfa_classify();
    rex_nfa_classify();
  }
   else if (action == "rex_stats") {
    // rank -arex_stats -zrexExprFile:TEST_REXs -zforced_len:20  -z(do_dfa|do_nfa|do_trx)
    extern void rex_stats();
    rex_stats();
  }
  // rank -affa_tests -zrexExprFile:<REX-in-file> -zrexInputFile:<string-input-file>
   else if (action == "ffa_tests") {
    extern void ffa_tests();
    ffa_tests();
  }
  else
    zop.usage_and_die("Unknown 'action' (-a) argument");
}

int main(int argc, char *argv[])
{
  //pin_processors();
  MyTime tm("main");
  get_options(argc, argv);
#if 1 //use this for testing
  init(argc, argv);
  main_action();
#else //use this +changes in dbg.h and MyTime.h for no debugging and timing info
  extern void rank_tests0();
  rank_tests0();
#endif
  return 0;
}
