#!/bin/bash
BASEDIR=$(dirname $0)
pushd $BASEDIR
BASE=.

if [[ $RANK == "" ]] ; then
    RANK=$BASE/../rank
fi

#EX1G=$BASE/../ex1G
EXBASE=$BASE/..

DATA=$BASE/DATA
OUT=$BASE/OUT
GOLD=$BASE/GOLD

if [ ! -d ${DATA} ] ; then
  echo "missing ${DATA}, probably running in the wrong directory"
  popd; exit 1;
fi

REXS=$(cd ${DATA}; ls *.rex)
GEXS="ex1G ex2G ex3G ex4G"

KINDS="do_dfa do_ffa do_nfa do_trx"
LENGTHS="10 100"


function run_single_test {
    local R=$1;#regular expression file
    local K=$2;#kind of ranking
    local L=$3;#max length
    local I=$4;#file with input strings
    local N=$5;#number of times to rank/unrank
    local COMB="$R-$K-$L-$I-$N"
    # shouldn't really blanket -zallow_ambiguous, but quick fix for now ...
    local CMD="${RANK} -aunit_test -zrexExprFile:$DATA/$R -zallow_ambiguous -zrexInputFile:$DATA/$I -zforced_len:$L -z$K -zrepeat:$N"
    #echo "CMD=$CMD"
    
    if ! $CMD 2>&1 > ${OUT}/${COMB}.out
    then
	if [ -f ${GOLD}/${COMB}.err ] ; then
	    echo "OK ... expected failure: $COMB: CMD=$CMD" | tee -a ${OUT}/errors
	else
	    echo "FAILED: $COMB: CMD=$CMD" | tee -a ${OUT}/errors
	fi
	mv ${OUT}/${COMB}.out ${OUT}/${COMB}.err
    else
	#grep gets confused by 'Timer:', although I try to ignore it ...
	grep -v Timer: ${OUT}/${COMB}.out > ${OUT}/${COMB}.out2
	rm ${OUT}/${COMB}.out
    fi
}

function run_grammar_test {
    local GEXE=$1
    local CMD=${EXBASE}/${GEXE}
    local GOUT=${OUT}/test_gram.${GEXE}.out
    local GERR=${OUT}/test_gram.${GEXE}.err
    local RES=""
    echo -n "Testing grammar ${GEXE} "
    
    if ! $CMD 2>&1 > ${GOUT}
    then
	echo "FAILED: CMD=$CMD" | tee -a ${OUT}/errors
	mv ${GOUT} ${OUT}/${GERR}
	RES="FAIL"
    else
	#grep gets confused by 'Timer:', although I try to ignore it ...
	grep -v Timer: ${GOUT} > ${GOUT}2
	rm ${GOUT}
    fi
    echo ${RES}
}

function run_tests {
    #if [ ! -d ${OUT} ] ; then mkdir ${OUT}; fi
    rm -rf ${OUT}
    mkdir ${OUT}
    
    for R in ${REXS} ; do
	for K in ${KINDS} ; do
	    PROBLEM=${R%.rex}.problematic
	    if [ -f ${DATA}/${PROBLEM} ] ; then
		echo "skip  problematic ";
	    else
	    	for L in ${LENGTHS} ; do
		    run_single_test $R $K $L in 10
		done
	    fi
	done
    done

    echo "" ; echo "GRAMMAR TESTS:"
    for EG in ${GEXS} ; do
	run_grammar_test ${EG};
    done

    echo""; echo "DIFF"
    diff -r -w --ignore-matching-lines='SVN_VERSION\|SVN_CHANGED\|Timer\|->unparse()=\|MemoryInfo' ${GOLD} ${OUT} -xLOG -xlog.diff-err -xerrors -x.svn 2>&1 | tee ${OUT}/log.diff-err
    if [[ ! -s ${OUT}/log.diff-err ]] ; then # i.e. if log.diff-err is not empty
	echo "PASSED"
    else
	echo "DIFFERENT"
    fi
}

run_tests;
popd >/dev/null; exit 0
