#ifndef _REGEX_RANKER_H
#define _REGEX_RANKER_H

#ifdef PYTHON_MODULE
#include <Python.h>
#else
#include "rankerobject.h"
#endif

// luchaupd
#include "luchaupd_wrapper.h"

#include <gmpxx.h>

enum regex_ranker_strategy {
    DFA_RANKING1 = 0,
    EMPTY1 = 1,
    DFA_RANKING2 = 2,
    EMPTY3 = 3,
    NFA_RANKING = 4
};

#ifdef PYTHON_MODULE
class regex_ranker {
# else
class regex_ranker : public rankerobject {
#endif
private:
    uint32_t _strategy;
    rankerobject * _ranker;
    std::string _regex;
    uint32_t _min_len;
    uint32_t _max_len;
    uint32_t _num_states;
    uint32_t _num_symbols;
public:
    regex_ranker(uint32_t, std::string, uint32_t);
    ~regex_ranker();
#ifdef PYTHON_MODULE
    std::string unrank(PyObject*);
    PyObject* rank(const std::string);
    PyObject* getNumWordsInLanguage();
    PyObject* getT(unsigned int, size_t);
#else
    std::string unrank(const mpz_class);
    mpz_class rank(const std::string);
    mpz_class getNumWordsInLanguage();
    mpz_class getT(unsigned int, size_t);
#endif
    uint32_t getNumStates();
    uint32_t getNumSymbols();
};

#endif /* _REGEX_RANKER_H */
