#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import time
import random
import unittest

sys.path.append('thirdparty/ffx')
sys.path.append('thirdparty/ranking-test-framework')

import FTE

TRIALS = 100

FORMATS = [
    ('(a|b)+b(a|b)+',16),
    ('(a|b)+b(a|b)+',32),
    ('abc(d|e|f)*def(d|e|f)*',16),
    ('abc(d|e|f)*def(d|e|f)*',32),
    ('[0-9]+', 16),
    ('[0-9]+', 17),
    ('[0-1]+', 16),
    ('[0-1]+', 17),
    ('[0-1]+', 128),
    ('[0-1]+', 1024),
    ('[0-9A-F]+', 4),
    ('[0-9A-F]+', 16),
    ('.+', 16),
    ('.+', 128),
]


class TestFFX(unittest.TestCase):

    def testFFXEncrypt1(self):
        K = '0' * 16
        for strategy in [0,2,4]:
            for input_format in FORMATS:
                for output_format in FORMATS:
                    print ['strategy='+str(strategy),'input=' +
                           str(input_format), 'ouput=' + str(output_format)]
                    try:
                        fpe = FTE.new(K,
                                      input_format=input_format,
                                      output_format=output_format,
                                      input_strategy=strategy,
                                      output_strategy=strategy)
                        for i in range(TRIALS):
                            N = random.randint(
                                0, fpe._input_ranker.getNumWordsInLanguage() - 1)
                            X = fpe._input_ranker.unrank(N)

                            start = time.time()
                            C = fpe.encrypt(X)
                            encrypt_cost = time.time() - start

                            start = time.time()
                            Y = fpe.decrypt(C)
                            decrypt_cost = time.time() - start
    
                            self.assertEquals(X, Y)
                    except FTE.OutputDomainTooSmallException:
                        continue
                    except FTE.EncryptionFailed:
                        continue


if __name__ == '__main__':
    unittest.main()
